/*
    m1cmds
    
    This plugins perform decoding video signal and descrambling command
    for VIDEOSTREAM equipment control.

    We try to use simples way for encoding data into bytes command.
    12-byte command consist of 3 4-byte + TAIL command should be sent to 
    VIDEOSTREAM equipment.
    
*/
    
#include <stdio.h>
#include "psvbi.h"
#include "psvbid-plugins.h"

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <syslog.h>
#include <pthread.h>

#define CORRECTIONS_MAX 5 /* maximum number of correction done in code */
#define SERIAL_DEV "/dev/ttyS0"
#define CMD_INTERVAL 2 /* 2 second is minimal commands interval */
#define CMD_RESPONSE 40 /* 40 ms is timeout for response */

static struct termios oldtio,newtio;
static int serial_handler;
static char* serial_device = SERIAL_DEV;
static struct timeval cmd_time;
static pthread_t thread_cmd_sender;

static void* cmd_sender(void* arg)
{
    // reply
    unsigned long reply;
    
    // prepare cmd
    unsigned long cmd = (unsigned long)arg;
    
    // async sending vars for
    fd_set fds;
    struct timeval tv;
    int r;
    struct timeval current_time;
    char current_time_str[255];
        
    // current time for console logging
    gettimeofday(&current_time, NULL);
    strftime(current_time_str, sizeof(current_time_str), "%Y-%m-%d %H:%M:%S", localtime(&current_time.tv_sec));
    printf("%s.%.6ld: ",current_time_str,current_time.tv_usec);

    // discard all data in buffers
    tcflush(serial_handler, TCIOFLUSH);
    
    // send cmd
    write(serial_handler, &cmd, sizeof(cmd));
    
    // async read
    FD_ZERO (&fds);
    FD_SET (serial_handler, &fds);
    tv.tv_sec = 0; /* Timeout 'sec'. */
    tv.tv_usec = CMD_RESPONSE*1000; /* Timeout 'usec'. */
    r = select (serial_handler + 1, &fds, NULL, NULL, &tv);

    // check if timeout/error occured
    if (r > 0)
    {
        // read data
        read(serial_handler, &reply, 2);

        // check for ack
        if (( reply & 0xFF) == 0x04)
        {
            // notify ACK
            syslog(LOG_NOTICE,"m1cmds: ACK[%d], CMD[%.8lX]", (reply>>8)&0xFF, cmd);
            printf("m1cmds: ACK[%d], CMD[%.8lX]\n", (reply>>8)&0xFF, cmd);
        }
        else
        if (( reply & 0xFF) == 0x05)
        {
            // notify NACK
            syslog(LOG_NOTICE,"m1cmds: NACK[%d], CMD[%.8lX]", (reply>>8)&0xFF, cmd);
            printf("m1cmds: NACK[%d], CMD[%.8lX]\n", (reply>>8)&0xFF, cmd);
        }
        else
        {
            // notify ERR
            syslog(LOG_NOTICE,"m1cmds: ERR[%.8lX], CMD[%.8lX]\n", reply, cmd);
            printf("m1cmds: ERR[%.8lX], CMD[%.8lX]\n", reply, cmd);
        }
    }
    else
    {
        // notify TIMEOUT
        syslog(LOG_NOTICE,"m1cmds: TIMEOUT[%d], CMD[%.8lX]", CMD_RESPONSE, cmd);
        printf("m1cmds: TIMEOUT[%d], CMD[%.8lX]\n", CMD_RESPONSE, cmd);
    }
    
    pthread_exit(NULL);
};


void __attribute__ ((constructor)) my_init(void)
{
    /* open syslog */
    /* openlog("m1cmd",LOG_PID , LOG_LOCAL6); */

    // open and configure device
    serial_handler = open(serial_device, O_RDWR | O_NOCTTY );
    if (serial_handler <0)
    {
        syslog(LOG_EMERG,"m1cmds: ERROR! Unable to open '%s'.",serial_device);
        fprintf(stderr, "m1cmds: ERROR! Unable to open '%s'.\n",serial_device);
        exit(-1);
    }
    else
    {
        syslog(LOG_NOTICE,"m1cmds: Device '%s' opened successfully",serial_device);
        printf("m1cmds: Device '%s' opened successfully\n",serial_device);
    }
        
    // configure serial device
    tcgetattr(serial_handler,&oldtio);        /* save current serial port settings */
    bzero(&newtio, sizeof(newtio));     /* clear struct for new port settings */
    newtio.c_iflag = IGNPAR;            /* ignore bytes with parity errors */
    newtio.c_oflag = 0;                 /* Raw output */
    newtio.c_lflag = 0;//ICANON;	/* enable canonical input */
    newtio.c_cc[VTIME]    = 0;		/* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;		/* blocking read until 5 chars received */
    newtio.c_cflag =
        B38400 |
        PARENB |
        PARODD |
        CS8 |
        CREAD |
        CLOCAL;
    tcflush(serial_handler, TCIFLUSH);
    tcsetattr(serial_handler,TCSANOW,&newtio);
    
    // fix init time
    gettimeofday(&cmd_time,NULL);
};

void __attribute__ ((destructor)) my_fini(void)
{
    /* check if device is present */
    if (serial_handler >0)
    {
        // restore serial device params
        tcflush(serial_handler, TCIFLUSH);
        tcsetattr(serial_handler,TCSANOW,&newtio);
    };
    
    /* notify about stop */
    syslog(LOG_NOTICE,"m1cmds: Unloading 'm1cmds.so'.");

    /* close log */
    /* closelog(); */
};


static int detect_ok(struct psvbi_block* block)
{
    char buf[1024];
    struct timeval t_current;
    unsigned long* cmd = (unsigned long*)block->decoded;

    // ignore datas with correction limit
    if(block->corrections > CORRECTIONS_MAX )
	return 1;

    // check for comand correctnes
    if((cmd[0] == cmd[1])&&(cmd[1] == cmd[2]))
	if((cmd[0] & 0xFF0000FF) == 0xFF000002)
	{
	    // command correct !!!
	    
	    // check how much time since last command retrieved
	    gettimeofday(&t_current, NULL);
	    if( (t_current.tv_sec - cmd_time.tv_sec) > CMD_INTERVAL )
	    {
	        // it's enouth period since last command
	        gettimeofday(&cmd_time,NULL);
	        
	        // start sending command asyncronously - in another thread
	        pthread_create(&thread_cmd_sender,NULL,cmd_sender,(void*)cmd[0]);
	    };

	    // return from handler
            return 1;	    
	};
/*	
    syslog(LOG_NOTICE, "m1cmds: ERROR! Incorrect command.");

    // create dump of data
    sprintf
    (
        buf,
        "m1cmds: CORRECTIONS=%d,POS=[%d,%d], DATA=[%.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X]",
        block->corrections,
        block->_h,block->_v,
        (unsigned int)block->decoded[0],(unsigned int)block->decoded[1],(unsigned int)block->decoded[2],
        (unsigned int)block->decoded[3],(unsigned int)block->decoded[4],(unsigned int)block->decoded[5],
        (unsigned int)block->decoded[6],(unsigned int)block->decoded[7],(unsigned int)block->decoded[8],
        (unsigned int)block->decoded[9],(unsigned int)block->decoded[10],(unsigned int)block->decoded[11],
        (unsigned int)block->decoded[12],(unsigned int)block->decoded[13]
    );
 
    syslog(LOG_NOTICE, buf);
*/
    
    return 0;
};


static int detect_fail(struct psvbi_block* block)
{
    return 0;
};


/* ---------------------------------------------------------------------------

    Define plugin attributest
    
----------------------------------------------------------------------------*/

struct psvbid_plugin PLUGIN = 
{
    .name = "m1cmds",
    .ok = detect_ok,
    .fail = detect_fail
};

