#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <dlfcn.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>              

#include "psvbid.h"
#include "psvbi.h"
#include "v4l2device.h"
#include "psvbid-plugins.h"

#define CLEAR(x) memset (&(x), 0, sizeof (x))

// static configuration
#define stat_frames_count 600*25 /* each 10 minutes we send info to log */
#define MAX_PLUGINS 10
#define VIDEO_DEV "/dev/video0"



// plugins working
static struct psvbid_plugin* plugins[MAX_PLUGINS];
static char* plugins_file[MAX_PLUGINS];
static long plugins_count = 0;

// statistic
static long stat_frames_processed = 0;
static long stat_frames_detected = 0;

// configuration params
static long detect_white_level = PSVBID_LEVEL_COMPARE;	/* base level of width to detect '1' */
static long detect_pos_x_start = 10;			/* defines rectange to search block */
static long detect_pos_x_stop = 30;
static long detect_pos_y_start = 0;
static long detect_pos_y_stop = 10;
static long demonize_flag = 0;

// init params
static char* video_dev = VIDEO_DEV;
static int video_dev_verbose = 0;

// handle ctrl-c
static sighandler_t sig_int;
static void int_handler(int signal)
{
    syslog(LOG_NOTICE,"Interrupt signal recieved!");
    video_stop();
};


// frame processor
static int psvbid_frame_process(int no, int index, void* data)
{
    
    psvbi_block block;
    CLEAR(block);

    int r = 0,i;

    // decode packet
    r = psvbi_find_Y
    (
      &block,
      data,
      detect_pos_x_start, detect_pos_x_stop,
      detect_pos_y_start, detect_pos_y_stop,
      PSVBID_FRAME_WIDTH,
      detect_white_level
    );

    // increment counters
    stat_frames_processed++;
    stat_frames_detected += (r != 0)?1:0;
    
    // enumerate plugins and call handlers
    for(i=0; i<plugins_count; i++)
    {
        if((r != 0)&&(plugins[i]->ok)) plugins[i]->ok(&block);
        if((r == 0)&&(plugins[i]->fail)) plugins[i]->fail(&block);
    };

    // report to syslog
    if( (stat_frames_processed % stat_frames_count) == (stat_frames_count - 1) )
    {
        // report to syslog
        syslog(LOG_INFO, "frames detected: %d (processed %d)", stat_frames_detected, stat_frames_processed );

        stat_frames_processed = 0;
        stat_frames_detected = 0;
    };
};

static void demonize(void)
{
    /* demonization */

    /* fork process */
    pid_t pid, sid;
    if ((pid = fork()) < 0)
    {
	perror("Error: Unable to fork");
        exit(EXIT_FAILURE);
    };
                                                                                                                                               
    /* exiting from parent */
    if (pid > 0)
    {
        exit(EXIT_SUCCESS);
    };
                                                                                                                                               
    /* changing file mode mask */
    umask(0);

                                                                                                                                               
    /* new sid */
    if ((sid = setsid()) < 0)
    {
	perror("Error: 'setsid'");
        exit(EXIT_FAILURE);
    };
                                                                                                                                               
    /* changing working directory */
    if (chdir("/") < 0)
    {
	perror("Error: Unable to change root directory");
        exit(EXIT_FAILURE);
    };
                                                                                                                                               
    /* closing standart file descriptors */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);

    // -----------------------------------------------------------------------
};

static void usage(void)
{
    printf
    (
	"Usage: psvbid <args>\n"
	"args:\n"
	"\t--demonize=N\t0-no daemon mode, 1-demonize (default %ld)\n"
	"\t--white-level=N\tWhite level of signal (default %ld)\n"
	"\t--pos-x-start=N\tleft 'x' position of search block (default %ld)\n"
	"\t--pos-x-stop=N\tright 'x' position of search block (default %ld)\n"
	"\t--pos-y-start=N\ttop 'y' position of search block (default %ld)\n"
	"\t--pos-y-stop=N\tbottom 'y' position of search block (default %ld)\n"
	"\t--video-dev=S\tvideo device to grab video from (default '%s')\n"
	"\t--load-plugin=S\tload handler plugin (multiple allowed, max %d)\n"
	"\t--verbose-video-init\tshow more info during init v4l2 device (default %d)\n"
	"\t--help\tShow program usage\n",
	demonize_flag,
	detect_white_level,
	detect_pos_x_start,
	detect_pos_x_stop,
	detect_pos_y_start,
	detect_pos_y_stop,
	video_dev,
	MAX_PLUGINS,
	video_dev_verbose
    );
};

static void used_params(void)
{
    printf
    (
	"Parameters list:\n"
	"\twhite-level=%ld\n"
	"\tpos-x-start=%ld\n"
	"\tpos-x-stop=%ld\n"
	"\tpos-y-start=%ld\n"
	"\tpos-y-stop=%ld\n"
	"\tvideo-dev=%s\n",
	detect_white_level,
	detect_pos_x_start,
	detect_pos_x_stop,
	detect_pos_y_start,
	detect_pos_y_stop,
	video_dev
    );
};



// options
static const char short_options [] = "d:w:x:a:y:b:v:p:ih";
static const struct option long_options [] = 
{
    { "demonize",	required_argument,	NULL,	'd'},
    { "white-level",	required_argument,	NULL,	'w'},
    { "pos-x-start",	required_argument,	NULL,	'x'},
    { "pos-x-stop",	required_argument,	NULL,	'a'},
    { "pos-y-start",	required_argument,	NULL,	'y'},
    { "pos-y-stop",	required_argument,	NULL,	'b'},
    { "video-dev",	required_argument,	NULL,	'v'},
    { "load-plugin",	required_argument,	NULL,	'p'},
    { "help",		no_argument,		NULL,	'h'},
    { "verbose-video-init", no_argument,	NULL,	'i'},
    { 0,		0,			0,	0}
};


int main(int argc, char** argv)
{
    int i,c,index;
    void *module, *plugin;
    char pwd[1024];
    char* pwd_tail;
    
    // save current path
    strncpy(pwd,argv[0],sizeof(pwd)-2);
    for(pwd_tail = pwd + strlen(pwd); (*pwd_tail != '/') && (pwd_tail != pwd); *pwd_tail-- = 0);
    /* printf("path='%s'\n",pwd); */

    // program arguments processing
    while(1)
    {
        c = getopt_long (argc, argv, short_options, long_options, &index);
        
        if(c == (-1)) break;
        
        switch(c)
        {
            case 0: break;
            case 'd': demonize_flag = atol(optarg);break;
            case 'w': detect_white_level = atol(optarg);break;
            case 'x': detect_pos_x_start = atol(optarg);break;
            case 'a': detect_pos_x_stop = atol(optarg);break;
            case 'y': detect_pos_y_start = atol(optarg);break;
            case 'b': detect_pos_y_stop = atol(optarg);break;
            case 'v': video_dev=optarg;break;
	    case 'i': video_dev_verbose = 1;break;
            case 'p': if(plugins_count<MAX_PLUGINS)plugins_file[plugins_count++] = optarg;break;
            case 'h': usage();exit(EXIT_FAILURE);break;
            default:
                printf("ERROR: Incorrect argument!\n");
                usage();
                exit (EXIT_FAILURE);
        };
    };

    // show parameters used
    used_params();

    // demonize
    if (demonize_flag)
    {
	demonize();
    };


    // load plugins
    fprintf(stderr,"Loading plugins [%d]:\n",plugins_count);
    for(i=0;i<plugins_count;i++)
    {
        // preparing filename
        char filename[1024];
        strcpy(filename,pwd);
        strcat(filename,"plugins/");
        strcat(filename,plugins_file[i]);
        strcat(filename,".so");
        printf("Loading '%s'\n", plugins_file[i]);

        // opening lib
        if(NULL == (module = dlopen(filename, RTLD_NOW)))
        {
            printf("ERROR: unable to load file: '%s'\n", filename);
            exit(EXIT_FAILURE);
        };
        
        // request plugin description struct
        if(NULL == (plugin = dlsym(module, "PLUGIN")))
        {
            printf("ERROR: unable to load 'PLUGIN' struct from file '%s'\n", filename);
            exit(EXIT_FAILURE);
        };
        
        // assigning
        plugins[i] = (struct psvbid_plugin*)plugin;
        plugins[i]->lib = module;
        
        // echo plugin name
        printf("SUCCESS: Plugin '%s' from file '%s' loaded!\n", plugins[i]->name, filename);
    };        

    // open video device
    video_open(video_dev_verbose,video_dev);

    
    /* open syslog */
    openlog("psvbid",LOG_PID, LOG_LOCAL6);

    // notify startup into log
    syslog(LOG_NOTICE,"Program started");
    
    // set new int handler
    sig_int = signal(SIGINT, int_handler);

    // start framegrabber
    video_start(psvbid_frame_process);

    /* report about transaction */
    syslog(LOG_NOTICE,"program finished");

    /* close log */
    closelog();

    // exit    
    exit(EXIT_SUCCESS);

    return 0;
};
